import copy
from pgameplayer import minimax_tree


# Runs minimax on Connect 4 game

# Yellow is the starting player

# PINF is +100

def count_two_in_row(board, player_token):
    '''

    :param board:
    :param player_token: string token
    :return:
    '''
    count = 0

    for r in range(6):
        for c in range(7):
            if r <= 4 and board[r][c] == board[r + 1][c] == player_token:
                count += 1
            if c <= 5 and board[r][c] == board[r][c + 1] == player_token:
                count += 1
            if r <= 4 and c <= 5 and board[r][c] == board[r + 1][c + 1] == player_token:
                count += 1
            if r >= 1 and c <= 5 and board[r][c] == board[r - 1][c + 1] == player_token:
                count += 1

    return count


def count_three_in_row(board, player_token):
    '''

    :param board:
    :param player_token: string token
    :return:
    '''
    count = 0

    for r in range(6):
        for c in range(7):
            if r <= 3 and board[r][c] == board[r + 1][c] == board[r + 2][c] == player_token:
                count += 1
            if c <= 4 and board[r][c] == board[r][c + 1] == board[r][c + 2] == player_token:
                count += 1
            if r <= 3 and c <= 4 and board[r][c] == board[r + 1][c + 1] == board[r + 2][c + 2] == player_token:
                count += 1
            if r >= 2 and c <= 4 and board[r][c] == board[r - 1][c + 1] == board[r - 2][c + 2] == player_token:
                count += 1

    return count


def connect_4_position_heuristic(board):
    '''

    :param board:
    :param player_token:
    :return:
    '''

    ry_diff = 0
    ry_diff += count_three_in_row(board, 'y')
    ry_diff += -count_three_in_row(board, 'r')
    return ry_diff


def win_for_player(board, player_token):
    """
    Four in a row, column or a diagonal
    :param board:
    :param player_token: 'r' / 'y'
    :return:
    """

    for r in range(6):
        for c in range(7):
            if board[r][c] == player_token and r <= 2:
                if board[r + 1][c] == board[r + 2][c] == board[r + 3][c] == player_token:
                    # vertical
                    return True

            if board[r][c] == player_token and c <= 3:
                if board[r][c + 1] == board[r][c + 2] == board[r][c + 3] == player_token:
                    # horizontal
                    return True

            if board[r][c] == player_token and c <= 3 and r <= 2:
                if board[r + 1][c + 1] == board[r + 2][c + 2] == board[r + 3][c + 3] == player_token:
                    # down diagonal
                    return True

            if board[r][c] == player_token and c <= 3 and r >= 3:
                if board[r - 1][c + 1] == board[r - 2][c + 2] == board[r - 3][c + 3] == player_token:
                    # up diagonal
                    return True

    return False


class Connect4Node(minimax_tree.Node):
    '''
        :param board: 3x3 character array with 'x','o' and '.'
    '''

    heuristic = connect_4_position_heuristic

    def __init__(self, board):
        self.state = board
        self.player = True
        self.value = None
        self.best_move = None

    def if_leaf(self):
        if win_for_player(self.state, 'r') or win_for_player(self.state, 'y'):
            return True

        if any('.' in row for row in self.state):
            return False

        return True

    def generate_moves(self, player):
        '''
        Generate possible moves for next player
        :return:
        '''

        curboard = copy.copy(self.state)

        next_state = []
        player = 'y' if player else 'r'

        for c in range(7):
            for r in range(6):
                if self.state[r][c] != '.':
                    newnode = Connect4Node(copy.deepcopy(self.state))
                    newnode.state[r][c] = player  # not following encapsulation
                    next_state.append(newnode)
        return next_state

    def evaluate(self):
        ''' Set value of board. If its not a win, loss or a draw then heuristic is evaluated.

        '''
        if win_for_player(self.state, 'x'):
            self.value = minimax_tree.PINF
        elif win_for_player(self.state, 'o'):
            self.value = minimax_tree.NINF
        elif not any('.' in row for row in self.state):
            self.value = 0
        else:
            self.value = Connect4Node.heuristic(self.state)

        # Connect 4 heuristic
        return self.value



import logging

logger = logging.getLogger("minimax")
# Straightforward minimax tree algorithm

PINF = 100
NINF = -100


class Node:
    ''' Node of minimax tree associated with a board state and player'''

    def __init__(self):
        self.state = None
        self.value = None
        self.player = True
        self.best_move = None

    def if_leaf(self):
        return True

    def generate_moves(self):
        return []

    def evaluate(self):
        return self.value

    def __str__(self):
        str = ''
        for r in range(len(self.state)):
            str += '|'.join(self.state[r]) + '\n'
        return str


def minimax(node, player):
    '''
    Main minimax function that obtains the best move to take
    :return:
    '''

    if node.if_leaf():
        return ([], node.evaluate())

    if player:
        maxv = NINF
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            (_, child.value) = minimax(child, not player)
            if child.value > maxv:
                maxv = child.value
                node.best_move = child.state

        node.value = maxv
        logger.debug("{} == {}".format(node.state, node.value))

        return (node.best_move, maxv)

    else:
        minv = PINF
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            (_, child.value) = minimax(child, not player)
            if child.value < minv:
                minv = child.value
                node.best_move = child.state
        node.value = minv
        logger.debug("{} == {}".format(node.state, node.value))

        return (node.best_move, minv)


# TODO: refactor to return next move too
def depth_limited_minimax(node, depth, player):
    '''
    Minimax algorithm that returns after a particular depth is reached
    :param node:
    :param depth:
    :param player:
    :return:
    '''
    if node.if_leaf() or depth == 0:
        return node.evaluate()

    if player:
        maxv = NINF
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            child.value = depth_limited_minimax(child, depth - 1, not player)
            if child.value > maxv:
                maxv = child.value
                node.best_move = child.state

        node.value = maxv
        logger.debug("{} == {}".format(node, node.value))
        return maxv

    else:
        minv = PINF
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            child.value = depth_limited_minimax(child, depth - 1, not player)
            if child.value < minv:
                minv = child.value
                node.best_move = child.state

        node.value = minv
        logger.debug("{} == {}".format(node, node.value))
        return minv


# alpha beta pruning takes a while

def alpha_beta_pruning_minimax(node, player, alpha, beta):
    '''
    Minimax variant that maintains max and min value for every node and prunes branches that are unnecessary
    '''

    if node.if_leaf():
        return node.evaluate()

    if player:
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            child.value = alpha_beta_pruning_minimax(child, not player, alpha, beta)
            if child.value > alpha:
                alpha = child.value
                node.best_move = child.state
            if alpha > beta:
                break

        node.value = alpha
        logger.debug(" {} == {}".format(node, node.value))
        return alpha

    else:
        possible_moves = node.generate_moves(player)
        for child in possible_moves:
            child.value = alpha_beta_pruning_minimax(child, not player, alpha, beta)
            if child.value < beta:
                beta = child.value
                node.best_move = child.state
            if alpha > beta:
                break
        node.value = beta
        logger.debug(" {} == {}".format(node, node.value))
        return beta



from pgameplayer import minimax_tree
import copy

# Making use of Minimax tree algorithm to solve board states of Tic Tac Toe


def compute_position_heuristic(board,player_token):
    '''

    :param board:
    :param player_token:
    :return:
    '''

    xo_diff = 0

    if board[0][0]=='x':
        xo_diff+=3
    elif board[0][0]=='o':
        xo_diff-=3

    if board[0][2]=='x':
        xo_diff+=3
    elif board[0][2] == 'o':
        xo_diff-=3

    if board[2][0]=='x':
        xo_diff+=3
    elif board[2][0] == 'o':
        xo_diff-=3

    if board[2][2]=='x':
        xo_diff+=3
    elif board[2][2] == 'o':
        xo_diff-=3

    if board[0][1]=='x':
        xo_diff+=1
    elif board[0][1] == 'o':
        xo_diff-=1

    if board[2][1]=='x':
        xo_diff+=1
    elif board[2][1] == 'o':
        xo_diff-=1

    if board[1][0]=='x':
        xo_diff+=3
    elif board[1][0] == 'o':
        xo_diff-=3

    if board[1][2]=='x':
        xo_diff+=3
    elif board[1][2] == 'o':
        xo_diff-=3

    if board[1][1]=='x':
        xo_diff+=5
    elif board[1][1] == 'o':
        xo_diff-=5

    return xo_diff

def compute_simple_heuristic(board,player_token):
    '''
    A heuristic that computes score of board states which are not leaves.
    :param board: 2d list
    :param player_token: bool
    :return:
    '''

    x_pot_wins = 0
    o_pot_wins = 0

    for r in range(3):
        if board[r][0] == board[r][1] == 'o' and board[r][2]== '.' or board[r][0] == board[r][2] == 'o' and board[r][1]== '.' \
                or board[r][2] == board[r][1] == 'o' and board[r][0]== '.':
            o_pot_wins += 1
    for r in range(3):
        if board[0][r] == board[1][r] == 'o' and board[2][r]== '.' or board[0][r] == board[2][r] == 'o' and board[1][r]== '.' \
                or board[2][r] == board[1][r] == 'o' and board[0][r]== '.':
            o_pot_wins += 1

    for r in range(3):
        if board[r][0] == board[r][1] == 'x' and board[r][2] == '.' or board[r][0] == board[r][2] == 'x' and board[r][
            1] == '.' \
                or board[r][2] == board[r][1] == 'x' and board[r][0] == '.':
            x_pot_wins += 1
    for r in range(3):
        if board[0][r] == board[1][r] == 'x' and board[2][r] == '.' or board[0][r] == board[2][r] == 'x' and board[1][
            r] == '.' \
                or board[2][r] == board[1][r] == 'x' and board[0][r] == '.':
            x_pot_wins += 1
    # diagonal check
    if board[0][0] == board[1][1] == 'x' and board[2][2] == '.' or board[0][0] == board[2][2] == 'x' and board[1][
        1] == '.' \
            or board[2][2] == board[1][1] == 'x' and board[0][0] == '.':
        x_pot_wins += 1

    if board[0][0] == board[1][1] == 'o' and board[2][2] == '.' or board[0][0] == board[2][2] == 'o' and board[1][
        1] == '.' \
            or board[2][2] == board[1][1] == 'o' and board[0][0] == '.':
        o_pot_wins += 1

    if board[0][2] == board[1][1] == 'x' and board[2][1] == '.' or board[0][2] == board[2][1] == 'x' and board[1][
        1] == '.' \
            or board[2][1] == board[1][1] == 'x' and board[0][2] == '.':
        x_pot_wins += 1

    if board[0][2] == board[1][1] == 'o' and board[2][1] == '.' or board[0][2] == board[2][1] == 'o' and board[1][
        1] == '.' \
            or board[2][1] == board[1][1] == 'o' and board[0][2] == '.':
        o_pot_wins += 1


    return (x_pot_wins - o_pot_wins)


def win_for_player(board,player_token):

    for r in range(3):
        if board[r][0]== player_token and board[r][1]== player_token and board[r][2]== player_token:
            return True
        if board[0][r] == player_token and board[1][r] == player_token and board[2][r] == player_token:
            return True

    # checking diagonal
    if board[0][0] == player_token and board[1][1] == player_token and board[2][2] == player_token:
        return True
    if board[0][2] == player_token and board[1][1] == player_token and board[2][0] == player_token:
        return True

    return False


class TicTacToeNode(minimax_tree.Node):
    '''
    Specialized class of Node.
    It represents one board position with the heuristic value for the given players move
    '''

    heuristic = compute_position_heuristic

    def __init__(self,board):
        '''

        :param board: 3x3 character array with 'x','o' and '.'
        '''
        self.state = board
        self.player = True
        self.value = None
        self.best_move = None


    def if_leaf(self):
        '''
        checks if node is either a win, loss or draw.
        :return: boolean
        '''

        if win_for_player(self.state,'x') or win_for_player(self.state,'o'):
            return True

        if any('.' in row for row in self.state):
            return False

        return True

    def generate_moves(self,player):
        '''
        Generates list of valid possible moves
        :param player: Boolean. x or o
        :return: list
        '''
        curboard = copy.copy(self.state)

        next_state = []
        if player:
            for r in range(3):
                for c in range(3):
                    if curboard[r][c] == '.':
                        newnode = TicTacToeNode(copy.deepcopy(self.state))
                        newnode.state[r][c] = 'x'         # not following encapsulation
                        next_state.append(newnode)
        else:
            for r in range(3):
                for c in range(3):
                    if curboard[r][c] == '.':
                        newnode = TicTacToeNode(copy.deepcopy(self.state))
                        newnode.state[r][c] = 'o'         # not following encapsulation
                        next_state.append(newnode)
        return next_state

    def evaluate(self):
        ''' Set value of board. If its not a win, loss or a draw then heuristic is evaluated.

        '''
        if win_for_player(self.state, 'x'):
            self.value = minimax_tree.PINF
        elif win_for_player(self.state,'o'):
            self.value = minimax_tree.NINF
        elif not any('.' in row for row in self.state):
            self.value = 0
        else:
            self.value = TicTacToeNode.heuristic(self.state,self.player)

        # XO heuristic
        return self.value


from setuptools import setup, find_packages

setup(name='perfect_game_player',
      version='1.2',
      description='Plays board games using Minimax algorithm',
      url='https://github.com/dsaw/perfect-game-player',
      author='Devesh Sawant',
      author_email='devesh47cool@gmail.com',
      packages=find_packages(exclude=['tests']),
      test_suite='nose.collector',
      tests_require=['nose'],
      license='GPLv3',
      install_requires=[],
      zip_safe=False
      )
